/* rxpop -- SAX2 parser driver for PyXML built on RXP.
*
* Copyright (c) 2002 Dave Kuhlman
* 
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#ifndef CHAR_SIZE
#define CHAR_SIZE 8
#endif

#include "system.h"
#include "ctype16.h"
#include "charset.h"
#include "string16.h"
#include "dtd.h"
#include "input.h"
#include "xmlparser.h"
#include "stdio16.h"
#include "version.h"
#include "namespaces.h"

#include <Python.h>

/* ------------------------------------------- */

static PyObject *ErrorObject;

/* ------------------------------------------- */

typedef struct _UserData {
	void * info;
	/* 
	RxpparseObject * info;
	*/
} UserData, * UserDataPtr;

/* ------------------------------------------- */

typedef void (*SAXFunc) (void * user_data, Parser parser, XBit pbit);

/* ------------------------------------------- */

typedef struct _SAXHandlers {
	SAXFunc startDocument;
	SAXFunc endDocument;
	SAXFunc startElement;
	SAXFunc endElement;
	SAXFunc startElementNS;
	SAXFunc endElementNS;
	SAXFunc characters;
	SAXFunc ignorableWhitespace;
	SAXFunc comment;
	SAXFunc cdataBlock;
	SAXFunc warning;
	SAXFunc error;
	SAXFunc fatalError;
	SAXFunc processingInstruction;
	SAXFunc notationDecl;
} SAXHandlers;

typedef struct _ParserFlags {
	int ExpandCharacterEntities;
	int ExpandGeneralEntities;
	int XMLSyntax;
	int XMLPredefinedEntities;
	int ErrorOnUnquotedAttributeValues;
	int NormaliseAttributeValues;
	int ErrorOnBadCharacterEntities;
	int ErrorOnUndefinedEntities;
	int ReturnComments;
	int CaseInsensitive;
	int ErrorOnUndefinedElements;
	int ErrorOnUndefinedAttributes;
	int WarnOnRedefinitions;
	int TrustSDD;
	int XMLExternalIDs;
	int ReturnDefaultedAttributes;
	int MergePCData;
	int XMLMiscWFErrors;
	int XMLStrictWFErrors;
	int AllowMultipleElements;
	int MaintainElementStack;
	int IgnoreEntities;
	int XMLLessThan;
	int IgnorePlacementErrors;
	int Validate;
	int ErrorOnValidityErrors;
	int XMLSpace;
	int XMLNamespaces;
	int NoNoDTDWarning;
	int SimpleErrorFormat;
	int AllowUndeclaredNSAttributes;
	int RelaxedAny;
	int ReturnNamespaceAttributes;
	int ProcessDTD;
} ParserFlags;

typedef struct _RxpparseObject {
	PyObject_HEAD
	PyObject * x_attr; /* Attributes dictionary */
	SAXHandlers handlers;
	Parser parser;
	ParserFlags parserFlags;
	PyObject * contentHandler;
	PyObject * dtdHandler;
	PyObject * errorHandler;
	
} RxpparseObject;

/* ------------------------------------------- */

static void startDocument(void * user_data, Parser parser, XBit pbit);
static void endDocument(void * user_data, Parser parser, XBit pbit);
static void startElement(void * user_data, Parser parser, XBit pbit);
static void endElement(void * user_data, Parser parser, XBit pbit);
static void startElementNS(void * user_data, Parser parser, XBit pbit);
static void endElementNS(void * user_data, Parser parser, XBit pbit);
static void characters(void * user_data, Parser parser, XBit pbit);
static void ignorableWhitespace(void * user_data, Parser parser, XBit pbit);
static void comment(void * user_data, Parser parser, XBit pbit);
static void cdataBlock(void * user_data, Parser parser, XBit pbit);
static void warning(void * user_data, Parser parser, XBit pbit);
static void error(void * user_data, Parser parser, XBit pbit);
static void fatalError(void * user_data, Parser parser, XBit pbit);
static void processingInstruction(void * user_data, Parser parser, XBit pbit);
static void notationDecl(void * user_data, Parser parser, XBit pbit);

static void processSource(RxpparseObject * user_data, Parser parser, InputSource source);
static void processBit(RxpparseObject * user_data, Parser parser, XBit pbit);
static void set_current_flags(RxpparseObject * obj);
static int get_flag(RxpparseObject * self, const char * flagName);
static int set_flag(RxpparseObject * self, const char * flagName, int flag);
static void setDefaultParserFlags(ParserFlags * parserFlags);
static int activate_methods(
	RxpparseObject * self,
	char * methodName,
	int flag
	);
static void register_callbacks(
	RxpparseObject * self,
	PyObject * handlerObj
	);
void formatError(Parser parser, XBit bit, char * pBuf);
static void warningCallback(XBit pbit, void * arg);

/* ------------------------------------------- */

static void setDefaultParserFlags(ParserFlags * parserFlags)
{
	parserFlags->XMLSyntax = 1;
	parserFlags->XMLPredefinedEntities = 1;
	parserFlags->XMLExternalIDs = 1;
	parserFlags->XMLMiscWFErrors = 1;
	parserFlags->ErrorOnUnquotedAttributeValues = 1;
	parserFlags->XMLLessThan = 1;
	parserFlags->ExpandGeneralEntities = 1;
	parserFlags->ExpandCharacterEntities = 1;
	parserFlags->NormaliseAttributeValues = 1;
	parserFlags->WarnOnRedefinitions = 1;
	parserFlags->TrustSDD = 1;
	parserFlags->ReturnComments = 1;
	parserFlags->MaintainElementStack = 1;
} /* setDefaultParserFlags */

/* ------------------------------------------- */

staticforward PyTypeObject Rxpparse_Type;

#define RxpparseObject_Check(v)	((v)->ob_type == &Rxpparse_Type)

static RxpparseObject *
newRxpparseObject(PyObject * pythonInstance)
{
	RxpparseObject * self;
	self = PyObject_New(RxpparseObject, &Rxpparse_Type);
	if (self == NULL)
	{
		return NULL;
	} /* if */
	self->x_attr = NULL;
	memset(&self->handlers, 0, sizeof(SAXHandlers));
	self->parser = NewParser();
	if (self->parser == NULL)
	{
		return NULL;
	} /* if */
	memset(&self->parserFlags, 0, sizeof(ParserFlags));
	setDefaultParserFlags(&self->parserFlags);
	self->contentHandler = NULL;
	self->dtdHandler = NULL;
	self->errorHandler = NULL;
	return self;
} /* newRxpparseObject */

/* ------------------------------------------- */

static void
Rxpparse_dealloc(RxpparseObject *self)
{
	Py_XDECREF(self->x_attr);
	PyObject_Del(self);
}

/* ------------------------------------------- */

#if 0


/* Invoke the Python callback.
*	Each C callback, calls the corresponding Python call-back,
	if there is one, and passes it the Python call-back object.
*/
static void setDocumentLocator(
	void *user_data,
	xmlSAXLocatorPtr loc
)
{
	RxpparseObject * info;
	PyObject * result;
	locatorobject * locatorObject;

	info = (RxpparseObject *)user_data;
	/* This test and the following one are redundant,
	*	but it might save a Python lookup.
	*/
	if ((info->handlers->setDocumentLocator != NULL) &&
		(PyObject_HasAttrString(info->pythonInstance, "setDocumentLocator")))
	{
		locatorObject = newlocatorobject(info->ctxt, loc);
		result = PyObject_CallMethod(info->pythonInstance, "setDocumentLocator",
			"O", locatorObject);
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* setDocumentLocator */

#endif


/* ------------------------------------------- */

/* Invoke the Python callback.
*	Each C callback, calls the corresponding Python call-back,
*	if there is one, and passes it the Python call-back object.
*/
static void startDocument(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	/* This test and the following one are redundant,
	*	but it might save a Python lookup.
	*/
	if ((info->handlers.startDocument != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "startDocument")))
	{
		result = PyObject_CallMethod(info->contentHandler, "startDocument",
			NULL);
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* startDocument */

/* ------------------------------------------- */

/* Invoke the Python callback.
*	Each C callback, calls the corresponding Python call-back,
*	if there is one, and passes it the Python call-back object.
*/
static void endDocument(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	/* This test and the following one are redundant,
	*	but it might save a Python lookup.
	*/
	if ((info->handlers.endDocument != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "endDocument")))
	{
		result = PyObject_CallMethod(info->contentHandler, "endDocument",
			NULL);
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* endDocument */

/* ------------------------------------------- */

static void startElement(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	const Char * name;
	Attribute attr;

	PyObject * attrDict;
	PyObject * result;
	PyObject * key;
	PyObject * value;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	if ((info->handlers.startElement != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "startElement")))
	{
		name = pbit->element_definition->name;
		attr = pbit->attributes;
		if (attr == NULL)
		{
			Py_INCREF(Py_None);
			attrDict = Py_None;
		}
		else
		{
			attrDict = PyDict_New();
			for (attr = pbit->attributes; attr; attr = attr->next)
			{
				key = PyString_FromString(attr->definition->name);
				if (attr->value != NULL)
				{
					value = PyString_FromString(attr->value);
				}
				else
				{
					Py_INCREF(Py_None);
					value = Py_None;
				} /* if */
				PyDict_SetItem(attrDict, key, value);
			} /* for */
		} /* if */

		result = PyObject_CallMethod(info->contentHandler, "startElement",
									 "sO", name, attrDict);
		Py_DECREF(attrDict);
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* startElement */

/* ------------------------------------------- */

static void endElement(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;
	const Char *name;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	if ((info->handlers.endElement != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "endElement")))
	{
		name = pbit->element_definition->name;
		result = PyObject_CallMethod(info->contentHandler, "endElement",
									 "s", name);
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* endElement */

/* ------------------------------------------- */

static void startElementNS(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	const Char * name;
	Attribute attr;

	PyObject * attrDict;
	PyObject * result;
	PyObject * key;
	PyObject * value;
	PyObject * pyTuple;
	PyObject * pyUri;
	PyObject * pyLocalname;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	if ((info->handlers.startElementNS != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "startElementNS")))
	{
		pyUri = PyString_FromString(
			pbit->ns_element_definition->namespace->uri);
		pyLocalname = PyString_FromString(pbit->element_definition->local);
		pyTuple = PyTuple_New(2);
		PyTuple_SetItem(pyTuple, 0, pyUri);
		PyTuple_SetItem(pyTuple, 1, pyLocalname);
		name = pbit->element_definition->name;
		attr = pbit->attributes;
		if (attr == NULL)
		{
			Py_INCREF(Py_None);
			attrDict = Py_None;
		}
		else
		{
			attrDict = PyDict_New();
			for (attr = pbit->attributes; attr; attr = attr->next)
			{
				key = PyString_FromString(attr->definition->name);
				if (attr->value != NULL)
				{
					value = PyString_FromString(attr->value);
				}
				else
				{
					Py_INCREF(Py_None);
					value = Py_None;
				} /* if */
				PyDict_SetItem(attrDict, key, value);
			} /* for */
		} /* if */

		result = PyObject_CallMethod(info->contentHandler, "startElementNS",
									 "OsO", pyTuple, name, attrDict);
		Py_DECREF(attrDict);
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* startElementNS */

/* ------------------------------------------- */

static void endElementNS(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;
	const Char *name;
	PyObject * pyTuple;
	PyObject * pyUri;
	PyObject * pyLocalname;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	if ((info->handlers.endElementNS != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "endElementNS")))
	{
		pyUri = PyString_FromString(
			pbit->ns_element_definition->namespace->uri);
		pyLocalname = PyString_FromString(pbit->element_definition->local);
		pyTuple = PyTuple_New(2);
		PyTuple_SetItem(pyTuple, 0, pyUri);
		PyTuple_SetItem(pyTuple, 1, pyLocalname);
		name = pbit->element_definition->name;
		result = PyObject_CallMethod(info->contentHandler, "endElementNS",
									 "Os", pyTuple, name);
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* endElementNS */

/* ------------------------------------------- */

static void characters(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;
	const Char * ch;
	char buf[1024];
	char * pBuf;
	int allocate;
	int len;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	ch = pbit->pcdata_chars;
	len = strlen(ch);
	if ((info->handlers.characters != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "characters")))
	{
		allocate = 0;
		if (len < 1024)
		{
			pBuf = buf;
		}
		else
		{
			pBuf = (char *)malloc(len + 1);
			allocate = 1;
		} /* if */
		strncpy(pBuf, ch, len);
		pBuf[len] = '\0';
		result = PyObject_CallMethod(info->contentHandler, "characters",
			"s", pBuf);
		if (allocate)
		{
			free(pBuf);
		} /* if */
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* characters */

/* ------------------------------------------- */

static void ignorableWhitespace(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;
	const Char * ch;
	char buf[1024];
	char * pBuf;
	int allocate;
	int len;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	ch = pbit->pcdata_chars;
	len = strlen(ch);
	if ((info->handlers.ignorableWhitespace != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "ignorableWhitespace")))
	{
		allocate = 0;
		if (len < 1024)
		{
			pBuf = buf;
		}
		else
		{
			pBuf = (char *)malloc(len + 1);
			allocate = 1;
		} /* if */
		strncpy(pBuf, ch, len);
		pBuf[len] = '\0';
		result = PyObject_CallMethod(info->contentHandler, "ignorableWhitespace",
			"s", pBuf);
		if (allocate)
		{
			free(pBuf);
		} /* if */
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* ignorableWhitespace */

/* ------------------------------------------- */

static void comment(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	const Char * value;
	PyObject * result;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	value = pbit->comment_chars;
	if ((info->handlers.comment != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "comment")))
	{
		result = PyObject_CallMethod(info->contentHandler, "comment",
			"s", value);
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* comment */

/* ------------------------------------------- */

static void cdataBlock(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	const Char * ch;
	int len;
	PyObject * result;
	char buf[1024];
	char * pBuf;
	int allocate;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	ch = pbit->cdsect_chars;
	if ((info->handlers.cdataBlock != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "cdataBlock")))
	{
		len = strlen(ch);
		allocate = 0;
		if (len < 1024)
		{
			pBuf = buf;
		}
		else
		{
			pBuf = (char *)malloc(len + 1);
			allocate = 1;
		} /* if */
		strncpy(pBuf, ch, len);
		pBuf[len] = '\0';
		result = PyObject_CallMethod(info->contentHandler, "cdataBlock",
			"s", pBuf);
		if (allocate)
		{
			free(pBuf);
		} /* if */
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* cdataBlock */

/* ------------------------------------------- */

void formatError(Parser parser, XBit bit, char * pBuf)
{
	int linenum, charnum;
	InputSource s, root;
	char buf[1024];

	pBuf[0] = '\0';
	root = ParserRootSource(parser);

	if(ParserGetFlag(parser, SimpleErrorFormat))
	{
		const char8 *d, *e;

		d = EntityDescription(root->entity);
		e = d+strlen8(d);
		while(e > d && e[-1] != '/')
		{
			--e;
		} /* while */

		if(parser->state == PS_validate_dtd)
		{
			sprintf(buf, "%s:-1(end of prolog):-1: ", e);
			strcat(pBuf, buf);
		}
		else if (parser->state == PS_validate_final)
		{
			sprintf(buf, "%s:-1(end of body):-1: ", e);
			strcat(pBuf, buf);
		}
		else
		{
			sprintf(buf, "%s:%d:%d: ", e,root->line_number+1, root->next+1);
			strcat(pBuf, buf);
		} /* if */

		if (bit->type == XBIT_warning)
		{
			sprintf(buf, "warning: ");
			strcat(pBuf, buf);
		} /* if */
		sprintf(buf, "%s\n", bit->error_message);
		strcat(pBuf, buf);

		return;
	} /* if */

	sprintf(buf, "%s: %s\n", 
			bit->type == XBIT_error ? "Error" : "Warning", 
			bit->error_message);
	strcat(pBuf, buf);

	if (parser->state == PS_validate_dtd || parser->state == PS_validate_final)
	{
		sprintf(buf, " (detected at end of %s of document %s)\n",
				parser->state == PS_validate_final ? "body" : "prolog",
				EntityDescription(root->entity));
		strcat(pBuf, buf);

		return;
	} /* if */

	for(s=parser->source; s; s=s->parent)
	{
		if(s->entity->name)
		{
			sprintf(buf, " in entity \"%s\"", s->entity->name);
			strcat(pBuf, buf);
		}
		else
		{
			sprintf(buf, " in unnamed entity");
			strcat(pBuf, buf);
		}

		switch(SourceLineAndChar(s, &linenum, &charnum))
		{
		case 1:
			sprintf(buf, " at line %d char %d of", linenum+1, charnum+1);
			strcat(pBuf, buf);
			break;
		case 0:
			sprintf(buf, " defined at line %d char %d of",
					linenum+1, charnum+1);
			strcat(pBuf, buf);
			break;
		case -1:
			sprintf(buf, " defined in");
			strcat(pBuf, buf);
			break;
		}

		sprintf(buf, " %s\n", EntityDescription(s->entity));
		strcat(pBuf, buf);
	} /* for */
} /* formatError */

/* ------------------------------------------- */

static void warning(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;
    char buf[1024];
	int processed;

	processed = 0;
	info = (RxpparseObject *)user_data;
	if (info->handlers.warning != NULL)
	{
		if (info->errorHandler != NULL)
		{
			if (PyObject_HasAttrString(info->errorHandler, "warning"))
			{
				processed = 1;
				formatError(parser, pbit, buf);
				result = PyObject_CallMethod(info->errorHandler, "warning",
											 "s", buf);
				if (PyErr_Occurred())
				{
					PyErr_Print();
				} /* if */
			} /* if */
		}
		else if (info->contentHandler != NULL)
		{
			if (PyObject_HasAttrString(info->contentHandler, "warning"))
			{
				processed = 1;
				formatError(parser, pbit, buf);
				result = PyObject_CallMethod(info->contentHandler, "warning",
											 "s", buf);
				if (PyErr_Occurred())
				{
					PyErr_Print();
				} /* if */
			} /* if */
		} /* if */
	} /* if */
		
	if (! processed)
	{
		ParserPerror(parser, pbit);
	} /* if */
} /* warning */

/* ------------------------------------------- */

static void warningCallback(XBit pbit, void * arg)
{
	RxpparseObject * info;
	
	info = (RxpparseObject *)arg;
	warning(arg, info->parser, pbit);
} /* warningCallback */

/* ------------------------------------------- */

static void error(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;
    char buf[1024];
	int processed;

	processed = 0;
	info = (RxpparseObject *)user_data;
	if (info->handlers.error != NULL)
	{
		if (info->errorHandler != NULL)
		{
			if (PyObject_HasAttrString(info->errorHandler, "error"))
			{
				formatError(parser, pbit, buf);
				result = PyObject_CallMethod(info->errorHandler, "error",
											 "s", buf);
				processed = 1;
				if (PyErr_Occurred())
				{
					PyErr_Print();
				} /* if */
			} /* if */
		} /* if */
		else if (info->contentHandler != NULL)
		{
			if (PyObject_HasAttrString(info->contentHandler, "error"))
			{
				formatError(parser, pbit, buf);
				result = PyObject_CallMethod(info->contentHandler, "error",
											 "s", buf);
				processed = 1;
				if (PyErr_Occurred())
				{
					PyErr_Print();
				} /* if */
			} /* if */
		} /* if */
	} /* if */

	if (! processed)
	{
		ParserPerror(parser, pbit);
	} /* if */
} /* error */

/* ------------------------------------------- */

static void fatalError(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;
    char buf[1024];
	int processed;

	processed = 0;
	info = (RxpparseObject *)user_data;
	if (info->handlers.fatalError != NULL)
	{
		if (info->errorHandler != NULL)
		{
			if (PyObject_HasAttrString(info->errorHandler, "fatalError"))
			{
				formatError(parser, pbit, buf);
				result = PyObject_CallMethod(info->errorHandler, "fatalError",
											 "s", buf);
				processed = 1;
				if (PyErr_Occurred())
				{
					PyErr_Print();
				} /* if */
			} /* if */
		} /* if */
		else if (info->contentHandler != NULL)
		{
			if (PyObject_HasAttrString(info->contentHandler, "fatalError"))
			{
				formatError(parser, pbit, buf);
				result = PyObject_CallMethod(info->contentHandler, "fatalError"
											 "s", buf);
				processed = 1;
				if (PyErr_Occurred())
				{
					PyErr_Print();
				} /* if */
			} /* if */
		} /* if */
	} /* if */

	if (! processed)
	{
		ParserPerror(parser, pbit);
	} /* if */
} /* fatalError */

/* ------------------------------------------- */

static void processingInstruction(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;
    char * target;
	char * data;

	info = (RxpparseObject *)user_data;
	if (info->contentHandler == NULL)
	{
		return;
	} /* if */
	if ((info->handlers.processingInstruction != NULL) &&
		(PyObject_HasAttrString(info->contentHandler, "processingInstruction")))
	{
		target = pbit->pi_name;
		data = pbit->pi_chars;
		result = PyObject_CallMethod(info->contentHandler, "processingInstruction",
			"ss", target, data);
		if (PyErr_Occurred())
		{
			PyErr_Print();
		} /* if */
	} /* if */
} /* processingInstruction */

/* ------------------------------------------- */

static void notationDecl(
	void * user_data,
	Parser parser,
	XBit pbit
)
{
	RxpparseObject * info;
	PyObject * result;
    const Char * name;
	const char8 * publicId;
	const char8 * systemId;

	info = (RxpparseObject *)user_data;
	if (info->dtdHandler == NULL)
	{
		return;
	} /* if */
	if ((info->handlers.notationDecl != NULL) &&
		(PyObject_HasAttrString(info->dtdHandler, "notationDecl")))
	{
		if ((parser->dtd != NULL) && (parser->dtd->notations != NULL))
		{
			name = parser->dtd->notations->name;
			publicId = parser->dtd->notations->publicid;
			systemId = parser->dtd->notations->systemid;
			result = PyObject_CallMethod(info->dtdHandler, "notationDecl",
										 "sss", name, publicId, systemId);
			if (PyErr_Occurred())
			{
				PyErr_Print();
			} /* if */
		} /* if */
	} /* if */
} /* notationDecl */

/* ------------------------------------------- */

static void
register_callbacks(
	RxpparseObject * self,
	PyObject * handlerObj
)
{
	int flag;
	SAXHandlers * saxHandler;

	saxHandler = &self->handlers;
	flag = PyObject_HasAttrString(handlerObj, "startDocument");
    if (flag)
	{
		saxHandler->startDocument = startDocument;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "endDocument");
    if (flag)
	{
		saxHandler->endDocument = endDocument;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "startElement");
    if (flag)
	{
		saxHandler->startElement = startElement;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "endElement");
    if (flag)
	{
		saxHandler->endElement = endElement;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "startElementNS");
    if (flag)
	{
		saxHandler->startElementNS = startElementNS;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "endElementNS");
    if (flag)
	{
		saxHandler->endElementNS = endElementNS;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "characters");
    if (flag)
	{
		saxHandler->characters = characters;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "ignorableWhitespace");
    if (flag)
	{
		saxHandler->ignorableWhitespace = ignorableWhitespace;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "processingInstruction");
    if (flag)
	{
		saxHandler->processingInstruction = processingInstruction;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "notationDecl");
    if (flag)
	{
		saxHandler->notationDecl = notationDecl;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "comment");
    if (flag)
	{
		saxHandler->comment = comment;
	} /* if */
#if 0
	flag = PyObject_HasAttrString(handlerObj, "warning");
    if (flag)
	{
		saxHandler->warning = warning;
		ParserSetWarningCallback(self->parser, warningCallback);
		ParserSetCallbackArg(self->parser, self);
	} /* if */
#endif
	flag = PyObject_HasAttrString(handlerObj, "error");
    if (flag)
	{
		saxHandler->error = error;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "fatalError");
    if (flag)
	{
		saxHandler->fatalError = fatalError;
	} /* if */
	flag = PyObject_HasAttrString(handlerObj, "cdataBlock");
    if (flag)
	{
		saxHandler->cdataBlock = cdataBlock;
	} /* if */

} /* register_callbacks */

/* ------------------------------------------- */

static void
processSource(RxpparseObject * user_data, Parser parser, InputSource source)
{
	XBit pbit;

	if(ParserPush(parser, source) == -1)
	{
		ParserPerror(parser, &parser->xbit);
		return;
	} /* if */

	startDocument(user_data, parser, NULL);
	while(1)
	{
		pbit = ReadXBit(parser);
		processBit(user_data, parser, pbit);
		if (pbit->type == XBIT_eof)
		{
			endDocument(user_data, parser, NULL);
			break;
		} /* if */
		if (pbit->type == XBIT_error)
		{
			return;
		} /* if */
	} /* while */
	FreeXBit(pbit);
} /* ProcessSource */

/* ------------------------------------------- */

static void processBit(RxpparseObject * user_data, Parser parser, XBit pbit)
{
	switch (pbit->type)
	{
	case XBIT_start:
	case XBIT_empty:
		if (user_data->parserFlags.XMLNamespaces &&
			pbit->ns_element_definition)
		{
			startElementNS(user_data, parser, pbit);
		}
		else
		{
			startElement(user_data, parser, pbit);
		} /* if */
		break;
	case XBIT_end:
		if (user_data->parserFlags.XMLNamespaces &&
			pbit->ns_element_definition)
		{
			endElementNS(user_data, parser, pbit);
		}
		else
		{
			endElement(user_data, parser, pbit);
		} /* if */
		break;
	case XBIT_pi:
		processingInstruction(user_data, parser, pbit);
		break;
	case XBIT_pcdata:
		if (pbit->i1)
		{
			ignorableWhitespace(user_data, parser, pbit);
		}
		else
		{
			characters(user_data, parser, pbit);
		} /* if */
		break;
	case XBIT_cdsect:
		cdataBlock(user_data, parser, pbit);
		break;
	case XBIT_dtd:
/*
		notationDecl(user_data, parser, pbit);
*/
		break;
	case XBIT_comment:
		comment(user_data, parser, pbit);
		break;
	case XBIT_error:
		error(user_data, parser, pbit);
		break;
	case XBIT_eof:
		break;
	default:
		fprintf(stderr, "\nUnknown event type %s\n", XBitTypeName[pbit->type]);
		exit(1);
		break;
	} /* switch */
} /* processBit */

/* ------------------------------------------- */

static void set_current_flags(RxpparseObject * obj)
{
	Parser p;
	ParserFlags * pf;

	p = obj->parser;
	pf = &obj->parserFlags;

	ParserSetFlag(p, ExpandCharacterEntities, pf->ExpandCharacterEntities);
	ParserSetFlag(p, ExpandGeneralEntities, pf->ExpandGeneralEntities);
	ParserSetFlag(p, XMLSyntax, pf->XMLSyntax);
	ParserSetFlag(p, XMLPredefinedEntities, pf->XMLPredefinedEntities);
	ParserSetFlag(p, ErrorOnUnquotedAttributeValues,
				  pf->ErrorOnUnquotedAttributeValues);
	ParserSetFlag(p, NormaliseAttributeValues, pf->NormaliseAttributeValues);
	ParserSetFlag(p, ErrorOnBadCharacterEntities,
				  pf->ErrorOnBadCharacterEntities);
	ParserSetFlag(p, ErrorOnUndefinedEntities, pf->ErrorOnUndefinedEntities);
	ParserSetFlag(p, ReturnComments, pf->ReturnComments);
	ParserSetFlag(p, CaseInsensitive, pf->CaseInsensitive);
	ParserSetFlag(p, ErrorOnUndefinedElements, pf->ErrorOnUndefinedElements);
	ParserSetFlag(p, ErrorOnUndefinedAttributes,
				  pf->ErrorOnUndefinedAttributes);
	ParserSetFlag(p, WarnOnRedefinitions, pf->WarnOnRedefinitions);
	ParserSetFlag(p, TrustSDD, pf->TrustSDD);
	ParserSetFlag(p, XMLExternalIDs, pf->XMLExternalIDs);
	ParserSetFlag(p, ReturnDefaultedAttributes, pf->ReturnDefaultedAttributes);
	ParserSetFlag(p, MergePCData, pf->MergePCData);
	ParserSetFlag(p, XMLMiscWFErrors, pf->XMLMiscWFErrors);
	ParserSetFlag(p, XMLStrictWFErrors, pf->XMLStrictWFErrors);
	ParserSetFlag(p, AllowMultipleElements, pf->AllowMultipleElements);
	ParserSetFlag(p, MaintainElementStack, pf->MaintainElementStack);
	ParserSetFlag(p, IgnoreEntities, pf->IgnoreEntities);
	ParserSetFlag(p, XMLLessThan, pf->XMLLessThan);
	ParserSetFlag(p, IgnorePlacementErrors, pf->IgnorePlacementErrors);
	ParserSetFlag(p, Validate, pf->Validate);
	ParserSetFlag(p, ErrorOnValidityErrors, pf->ErrorOnValidityErrors);
	ParserSetFlag(p, XMLSpace, pf->XMLSpace);
	ParserSetFlag(p, XMLNamespaces, pf->XMLNamespaces);
	ParserSetFlag(p, NoNoDTDWarning, pf->NoNoDTDWarning);
	ParserSetFlag(p, SimpleErrorFormat, pf->SimpleErrorFormat);
	ParserSetFlag(p, AllowUndeclaredNSAttributes,
				  pf->AllowUndeclaredNSAttributes);
	ParserSetFlag(p, RelaxedAny, pf->RelaxedAny);
	ParserSetFlag(p, ReturnNamespaceAttributes, pf->ReturnNamespaceAttributes);
	ParserSetFlag(p, ProcessDTD, pf->ProcessDTD);
} /* set_current_flags */

/* ------------------------------------------- */

static char Rxpparse_parse__doc__[] = "\
Parse a file, delivering events to event handler methods in the handler.\n\
Usage:\n\
    handler = MyHandler()\n\
    parser = rxpop.ParserCreate(handler)\n\
    parser.parse('sample.xml')\n\
";
static PyObject *
Rxpparse_parse(RxpparseObject * self, PyObject * args)
{
	char * fileName;
	Entity ent;
	InputSource source;
	int result;
	int flag;
	PyObject * pyResult;

	result = 0;
	if (!PyArg_ParseTuple(args, "s:parse", &fileName))
	{
		return NULL;
	} /* if */

	ent = NewExternalEntity(0, 0, fileName, 0, 0);
	if (ent)
	{
		self->parser = NewParser();

		if (self->errorHandler != NULL)
		{
			flag = PyObject_HasAttrString(self->errorHandler, "warning");
			if (flag)
			{
				self->handlers.warning = warning;
				ParserSetWarningCallback(self->parser, warningCallback);
				ParserSetCallbackArg(self->parser, self);
			} /* if */
		}
		else if (self->contentHandler != NULL)
		{
			flag = PyObject_HasAttrString(self->contentHandler, "warning");
			if (flag)
			{
				self->handlers.warning = warning;
				ParserSetWarningCallback(self->parser, warningCallback);
				ParserSetCallbackArg(self->parser, self);
			} /* if */
		} /* if */

		set_current_flags(self);
		source = EntityOpen(ent);
		processSource(self, self->parser, source);
		FreeDtd(self->parser->dtd);
		FreeParser(self->parser);
		self->parser = NULL;
	}

	if (result != 0)
	{
		pyResult = PyInt_FromLong((long)0);
		return pyResult;
	} /* if */

	pyResult = PyInt_FromLong((long)1);
	return pyResult;
} /* Rxpparse_parse */

/* ------------------------------------------- */



static char Rxpparse_parseFile__doc__[] = "\
Parse a file, delivering events to event handler methods in the handler.\n\
Usage:\n\
    handler = MyHandler()\n\
    parser = rxpop.ParserCreate(handler)\n\
    parser.parseFile('sample.xml')\n\
";
static PyObject *
Rxpparse_parseFile(RxpparseObject *self, PyObject *args)
{
	char * fileName;
	Entity ent;
	InputSource source;
	int result;
	PyObject * pyResult;

	result = 0;
	if (!PyArg_ParseTuple(args, "s:parse", &fileName))
	{
		return NULL;
	} /* if */

	ent = NewExternalEntity(0, 0, fileName, 0, 0);
	if (ent)
	{
		self->parser = NewParser();
		set_current_flags(self);
		source = EntityOpen(ent);
		processSource(self, self->parser, source);
		FreeDtd(self->parser->dtd);
		FreeParser(self->parser);
		self->parser = NULL;
	}

	if (result != 0)
	{
		pyResult = PyInt_FromLong((long)0);
		return pyResult;
	} /* if */

	pyResult = PyInt_FromLong((long)1);
	return pyResult;
} /* Rxpparse_parseFile */

/* ------------------------------------------- */

static char Rxpparse_setContentHandler__doc__[] = "\
Create a new Parser object.\n\
";
static PyObject *
Rxpparse_setContentHandler(RxpparseObject *self, PyObject *args)
{
	PyObject * pythonInstance;

	if (!PyArg_ParseTuple(args, "O:setContentHandler", &pythonInstance))
	{
		return NULL;
	} /* if */
	if (! PyInstance_Check(pythonInstance))
	{
		PyErr_SetString(PyExc_TypeError, "handler instance required");
		return NULL;
	} /* if */
	self->contentHandler = pythonInstance;
	register_callbacks(self, pythonInstance);
	Py_INCREF(Py_None);
	return Py_None;
} /* Rxpparse_setContentHandler */

/* ------------------------------------------- */

static char Rxpparse_setDTDHandler__doc__[] = "\
Create a new Parser object.\n\
";
static PyObject *
Rxpparse_setDTDHandler(RxpparseObject * self, PyObject *args)
{
	PyObject * pythonInstance;

	if (!PyArg_ParseTuple(args, "O:setDTDHandler", &pythonInstance))
	{
		return NULL;
	} /* if */
	if (! PyInstance_Check(pythonInstance))
	{
		PyErr_SetString(PyExc_TypeError, "handler instance required");
		return NULL;
	} /* if */
	self->dtdHandler = pythonInstance;
	register_callbacks(self, pythonInstance);
	Py_INCREF(Py_None);
	return Py_None;
} /* Rxpparse_setDTDHandler */

/* ------------------------------------------- */

static char Rxpparse_setErrorHandler__doc__[] = "\
Create a new Parser object.\n\
";
static PyObject *
Rxpparse_setErrorHandler(RxpparseObject * self, PyObject *args)
{
	PyObject * pythonInstance;

	if (!PyArg_ParseTuple(args, "O:setErrorHandler", &pythonInstance))
	{
		return NULL;
	} /* if */
	if (! PyInstance_Check(pythonInstance))
	{
		PyErr_SetString(PyExc_TypeError, "handler instance required");
		return NULL;
	} /* if */
	self->errorHandler = pythonInstance;
	register_callbacks(self, pythonInstance);
	Py_INCREF(Py_None);
	return Py_None;
} /* Rxpparse_setErrorHandler */

/* ------------------------------------------- */

static char Rxpparse_getLocation__doc__[] = "\
Return the current location.\n\
Usage:\n\
    line,column = parser.getLocation()\n\
Returns a 2-tuple containing line number and column number.\n\
";
static PyObject *
Rxpparse_getLocation(RxpparseObject * self, PyObject *args)
{
	int linenum;
	int charnum;
	int result;
	PyObject * pyTuple;
	PyObject * pyLinenum;
	PyObject * pyCharnum;

	if (!PyArg_ParseTuple(args, ":getLocation"))
	{
		return NULL;
	} /* if */

	result = SourceLineAndChar(self->parser->source, &linenum, &charnum);
	if (result >= 0)
	{
		linenum++;
	}
	else
	{
		linenum = -1;
		charnum = -1;
	} /* if */
	pyLinenum = PyInt_FromLong((long)linenum);
	pyCharnum = PyInt_FromLong((long)charnum);
	pyTuple = PyTuple_New(2);
	PyTuple_SetItem(pyTuple, 0, pyLinenum);
	PyTuple_SetItem(pyTuple, 1, pyCharnum);
	return pyTuple;
} /* Rxpparse_getLocation */

/* ------------------------------------------- */

static int activate_methods(
	RxpparseObject * self,
	char * methodName,
	int flag
)
{
	SAXHandlers * handlers;

	handlers = &self->handlers;
	if (strcmp(methodName, "startDocument") == 0)
	{
		if (flag)
		{
			handlers->startDocument = startDocument;
		}
		else
		{
			handlers->startDocument = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "endDocument") == 0)
	{
		if (flag)
		{
			handlers->endDocument = endDocument;
		}
		else
		{
			handlers->endDocument = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "startElement") == 0)
	{
		if (flag)
		{
			handlers->startElement = startElement;
		}
		else
		{
			handlers->startElement = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "endElement") == 0)
	{
		if (flag)
		{
			handlers->endElement = endElement;
		}
		else
		{
			handlers->endElement = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "startElementNS") == 0)
	{
		if (flag)
		{
			handlers->startElementNS = startElementNS;
		}
		else
		{
			handlers->startElementNS = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "endElementNS") == 0)
	{
		if (flag)
		{
			handlers->endElementNS = endElementNS;
		}
		else
		{
			handlers->endElementNS = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "characters") == 0)
	{
		if (flag)
		{
			handlers->characters = characters;
		}
		else
		{
			handlers->characters = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "ignorableWhitespace") == 0)
	{
		if (flag)
		{
			handlers->ignorableWhitespace = ignorableWhitespace;
		}
		else
		{
			handlers->ignorableWhitespace = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "processingInstruction") == 0)
	{
		if (flag)
		{
			handlers->processingInstruction = processingInstruction;
		}
		else
		{
			handlers->processingInstruction = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "notationDecl") == 0)
	{
		if (flag)
		{
			handlers->notationDecl = notationDecl;
		}
		else
		{
			handlers->notationDecl = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "comment") == 0)
	{
		if (flag)
		{
			handlers->comment = comment;
		}
		else
		{
			handlers->comment = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "warning") == 0)
	{
		if (flag)
		{
			handlers->warning = warning;
			ParserSetWarningCallback(self->parser, warningCallback);
			ParserSetCallbackArg(self->parser, self);
		}
		else
		{
			handlers->warning = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "error") == 0)
	{
		if (flag)
		{
			handlers->error = error;
		}
		else
		{
			handlers->error = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "fatalError") == 0)
	{
		if (flag)
		{
			handlers->fatalError = fatalError;
		}
		else
		{
			handlers->fatalError = NULL;
		} /* if */
	}
	else if (strcmp(methodName, "cdataBlock") == 0)
	{
		if (flag)
		{
			handlers->cdataBlock = cdataBlock;
		}
		else
		{
			handlers->cdataBlock = NULL;
		} /* if */
	} /* if */

	return 1;
} /* activate_methods */

/* ------------------------------------------- */

static char Rxpparse_activate__doc__[] = "\
Activate (de-activate) an event method.\n\
Usage:\n\
    parser.activate(handlerName, flag)\n\
Where:\n\
    handlerName is 'startDocument', 'endDocument', 'characters', etc.\n\
    flag is 1 (activate) or 2 (de-activate)\n\
";
static PyObject *
Rxpparse_activate(RxpparseObject *self, PyObject *args)
{
	char * methodName;
	int flag;
	int result;
	PyObject * pyResult;

	if (!PyArg_ParseTuple(args, "si:activate", &methodName, &flag))
	{
		return NULL;
	} /* if */

	result = activate_methods(self, methodName, flag);
	pyResult = PyInt_FromLong((long)0);
	return pyResult;
} /* Rxpparse_activate */

/* ------------------------------------------- */

static int get_flag(RxpparseObject * self, const char * flagName)
{
	int flag;

	flag = 0;

	if (strcmp(flagName, "ExpandCharacterEntities") == 0)
	{
		flag = ParserGetFlag(self->parser, ExpandCharacterEntities);
	}
	else if (strcmp(flagName, "ExpandGeneralEntities") == 0)
	{
		flag = ParserGetFlag(self->parser, ExpandGeneralEntities);
	}
	else if (strcmp(flagName, "XMLSyntax") == 0)
	{
		flag = ParserGetFlag(self->parser, XMLSyntax);
	}
	else if (strcmp(flagName, "XMLPredefinedEntities") == 0)
	{
		flag = ParserGetFlag(self->parser, XMLPredefinedEntities);
	}
	else if (strcmp(flagName, "ErrorOnUnquotedAttributeValues") == 0)
	{
		flag = ParserGetFlag(self->parser, ErrorOnUnquotedAttributeValues);
	}
	else if (strcmp(flagName, "NormaliseAttributeValues") == 0)
	{
		flag = ParserGetFlag(self->parser, NormaliseAttributeValues);
	}
	else if (strcmp(flagName, "ErrorOnBadCharacterEntities") == 0)
	{
		flag = ParserGetFlag(self->parser, ErrorOnBadCharacterEntities);
	}
	else if (strcmp(flagName, "ErrorOnUndefinedEntities") == 0)
	{
		flag = ParserGetFlag(self->parser, ErrorOnUndefinedEntities);
	}
	else if (strcmp(flagName, "ReturnComments") == 0)
	{
		flag = ParserGetFlag(self->parser, ReturnComments);
	}
	else if (strcmp(flagName, "CaseInsensitive") == 0)
	{
		flag = ParserGetFlag(self->parser, CaseInsensitive);
	}
	else if (strcmp(flagName, "ErrorOnUndefinedElements") == 0)
	{
		flag = ParserGetFlag(self->parser, ErrorOnUndefinedElements);
	}
	else if (strcmp(flagName, "ErrorOnUndefinedAttributes") == 0)
	{
		flag = ParserGetFlag(self->parser, ErrorOnUndefinedAttributes);
	}
	else if (strcmp(flagName, "WarnOnRedefinitions") == 0)
	{
		flag = ParserGetFlag(self->parser, WarnOnRedefinitions);
	}
	else if (strcmp(flagName, "TrustSDD") == 0)
	{
		flag = ParserGetFlag(self->parser, TrustSDD);
	}
	else if (strcmp(flagName, "XMLExternalIDs") == 0)
	{
		flag = ParserGetFlag(self->parser, XMLExternalIDs);
	}
	else if (strcmp(flagName, "ReturnDefaultedAttributes") == 0)
	{
		flag = ParserGetFlag(self->parser, ReturnDefaultedAttributes);
	}
	else if (strcmp(flagName, "MergePCData") == 0)
	{
		flag = ParserGetFlag(self->parser, MergePCData);
	}
	else if (strcmp(flagName, "XMLMiscWFErrors") == 0)
	{
		flag = ParserGetFlag(self->parser, XMLMiscWFErrors);
	}
	else if (strcmp(flagName, "XMLStrictWFErrors") == 0)
	{
		flag = ParserGetFlag(self->parser, XMLStrictWFErrors);
	}
	else if (strcmp(flagName, "AllowMultipleElements") == 0)
	{
		flag = ParserGetFlag(self->parser, AllowMultipleElements);
	}
	else if (strcmp(flagName, "MaintainElementStack") == 0)
	{
		flag = ParserGetFlag(self->parser, MaintainElementStack);
	}
	else if (strcmp(flagName, "IgnoreEntities") == 0)
	{
		flag = ParserGetFlag(self->parser, IgnoreEntities);
	}
	else if (strcmp(flagName, "XMLLessThan") == 0)
	{
		flag = ParserGetFlag(self->parser, XMLLessThan);
	}
	else if (strcmp(flagName, "IgnorePlacementErrors") == 0)
	{
		flag = ParserGetFlag(self->parser, IgnorePlacementErrors);
	}
	else if (strcmp(flagName, "Validate") == 0)
	{
		flag = ParserGetFlag(self->parser, Validate);
	}
	else if (strcmp(flagName, "ErrorOnValidityErrors") == 0)
	{
		flag = ParserGetFlag(self->parser, ErrorOnValidityErrors);
	}
	else if (strcmp(flagName, "XMLSpace") == 0)
	{
		flag = ParserGetFlag(self->parser, XMLSpace);
	}
	else if (strcmp(flagName, "XMLNamespaces") == 0)
	{
		flag = ParserGetFlag(self->parser, XMLNamespaces);
	}
	else if (strcmp(flagName, "NoNoDTDWarning") == 0)
	{
		flag = ParserGetFlag(self->parser, NoNoDTDWarning);
	}
	else if (strcmp(flagName, "SimpleErrorFormat") == 0)
	{
		flag = ParserGetFlag(self->parser, SimpleErrorFormat);
	}
	else if (strcmp(flagName, "AllowUndeclaredNSAttributes") == 0)
	{
		flag = ParserGetFlag(self->parser, AllowUndeclaredNSAttributes);
	}
	else if (strcmp(flagName, "RelaxedAny") == 0)
	{
		flag = ParserGetFlag(self->parser, RelaxedAny);
	}
	else if (strcmp(flagName, "ReturnNamespaceAttributes") == 0)
	{
		flag = ParserGetFlag(self->parser, ReturnNamespaceAttributes);
	}
	else if (strcmp(flagName, "ProcessDTD") == 0)
	{
		flag = ParserGetFlag(self->parser, ProcessDTD);
	} /* if */

	return flag;
} /* get_flag */

/* ------------------------------------------- */

static char Rxpparse_getParserFlag__doc__[] = "\
Activate (de-activate) an event method.\n\
Usage:\n\
    parser.getParserFlag(flagName, flag)\n\
Where:\n\
    flagName is one of the following:\n\
        ExpandCharacterEntities\n\
        ExpandGeneralEntities\n\
        XMLSyntax\n\
        XMLPredefinedEntities\n\
        ErrorOnUnquotedAttributeValues\n\
        NormaliseAttributeValues\n\
        ErrorOnBadCharacterEntities\n\
        ErrorOnUndefinedEntities\n\
        ReturnComments\n\
        CaseInsensitive\n\
        ErrorOnUndefinedElements\n\
        ErrorOnUndefinedAttributes\n\
        WarnOnRedefinitions\n\
        TrustSDD\n\
        XMLExternalIDs\n\
        ReturnDefaultedAttributes\n\
        MergePCData\n\
        XMLMiscWFErrors\n\
        XMLStrictWFErrors\n\
        AllowMultipleElements\n\
        MaintainElementStack\n\
        IgnoreEntities\n\
        XMLLessThan\n\
        IgnorePlacementErrors\n\
        Validate\n\
        ErrorOnValidityErrors\n\
        XMLSpace\n\
        XMLNamespaces\n\
        NoNoDTDWarning\n\
        SimpleErrorFormat\n\
        AllowUndeclaredNSAttributes\n\
        RelaxedAny\n\
        ReturnNamespaceAttributes\n\
        ProcessDTD\n\
    flag is 0 (clear) or 1 (set)\n\
Returns: 1 if successful.  0 if unsuccessful (e.g. no such flag).\n\
";
static PyObject *
Rxpparse_getParserFlag(RxpparseObject * self, PyObject *args)
{
	char * flagName;
	int flag;
	PyObject * pyFlag;

	if (!PyArg_ParseTuple(args, "s:getParserFlag", &flagName))
	{
		return NULL;
	} /* if */

	flag = get_flag(self, flagName);
	if (flag != 0)
	{
		flag = 1;
	} /* if */
	pyFlag = PyInt_FromLong((long)flag);
	return pyFlag;
} /* Rxpparse_getParserFlag */

/* ------------------------------------------- */

static int set_flag(RxpparseObject * self, const char * flagName, int flag)
{
	int result;

	result = 0;
	if (flag != 0)
	{
		flag = 1;
	} /* if */

	if (strcmp(flagName, "ExpandCharacterEntities") == 0)
	{
		ParserSetFlag(self->parser, ExpandCharacterEntities, flag);
		self->parserFlags.ExpandCharacterEntities = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ExpandGeneralEntities") == 0)
	{
		ParserSetFlag(self->parser, ExpandGeneralEntities, flag);
		self->parserFlags.ExpandGeneralEntities = flag;
		result = 1;
	}
	else if (strcmp(flagName, "XMLSyntax") == 0)
	{
		ParserSetFlag(self->parser, XMLSyntax, flag);
		self->parserFlags.XMLSyntax = flag;
		result = 1;
	}
	else if (strcmp(flagName, "XMLPredefinedEntities") == 0)
	{
		ParserSetFlag(self->parser, XMLPredefinedEntities, flag);
		self->parserFlags.XMLPredefinedEntities = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ErrorOnUnquotedAttributeValues") == 0)
	{
		ParserSetFlag(self->parser, ErrorOnUnquotedAttributeValues, flag);
		self->parserFlags.ErrorOnUnquotedAttributeValues = flag;
		result = 1;
	}
	else if (strcmp(flagName, "NormaliseAttributeValues") == 0)
	{
		ParserSetFlag(self->parser, NormaliseAttributeValues, flag);
		self->parserFlags.NormaliseAttributeValues = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ErrorOnBadCharacterEntities") == 0)
	{
		ParserSetFlag(self->parser, ErrorOnBadCharacterEntities, flag);
		self->parserFlags.ErrorOnBadCharacterEntities = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ErrorOnUndefinedEntities") == 0)
	{
		ParserSetFlag(self->parser, ErrorOnUndefinedEntities, flag);
		self->parserFlags.ErrorOnUndefinedEntities = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ReturnComments") == 0)
	{
		ParserSetFlag(self->parser, ReturnComments, flag);
		self->parserFlags.ReturnComments = flag;
		result = 1;
	}
	else if (strcmp(flagName, "CaseInsensitive") == 0)
	{
		ParserSetFlag(self->parser, CaseInsensitive, flag);
		self->parserFlags.CaseInsensitive = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ErrorOnUndefinedElements") == 0)
	{
		ParserSetFlag(self->parser, ErrorOnUndefinedElements, flag);
		self->parserFlags.ErrorOnUndefinedElements = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ErrorOnUndefinedAttributes") == 0)
	{
		ParserSetFlag(self->parser, ErrorOnUndefinedAttributes, flag);
		self->parserFlags.ErrorOnUndefinedAttributes = flag;
		result = 1;
	}
	else if (strcmp(flagName, "WarnOnRedefinitions") == 0)
	{
		ParserSetFlag(self->parser, WarnOnRedefinitions, flag);
		self->parserFlags.WarnOnRedefinitions = flag;
		result = 1;
	}
	else if (strcmp(flagName, "TrustSDD") == 0)
	{
		ParserSetFlag(self->parser, TrustSDD, flag);
		self->parserFlags.TrustSDD = flag;
		result = 1;
	}
	else if (strcmp(flagName, "XMLExternalIDs") == 0)
	{
		ParserSetFlag(self->parser, XMLExternalIDs, flag);
		self->parserFlags.XMLExternalIDs = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ReturnDefaultedAttributes") == 0)
	{
		ParserSetFlag(self->parser, ReturnDefaultedAttributes, flag);
		self->parserFlags.ReturnDefaultedAttributes = flag;
		result = 1;
	}
	else if (strcmp(flagName, "MergePCData") == 0)
	{
		ParserSetFlag(self->parser, MergePCData, flag);
		self->parserFlags.MergePCData = flag;
		result = 1;
	}
	else if (strcmp(flagName, "XMLMiscWFErrors") == 0)
	{
		ParserSetFlag(self->parser, XMLMiscWFErrors, flag);
		self->parserFlags.XMLMiscWFErrors = flag;
		result = 1;
	}
	else if (strcmp(flagName, "XMLStrictWFErrors") == 0)
	{
		ParserSetFlag(self->parser, XMLStrictWFErrors, flag);
		self->parserFlags.XMLStrictWFErrors = flag;
		result = 1;
	}
	else if (strcmp(flagName, "AllowMultipleElements") == 0)
	{
		ParserSetFlag(self->parser, AllowMultipleElements, flag);
		self->parserFlags.AllowMultipleElements = flag;
		result = 1;
	}
	else if (strcmp(flagName, "MaintainElementStack") == 0)
	{
		ParserSetFlag(self->parser, MaintainElementStack, flag);
		self->parserFlags.MaintainElementStack = flag;
		result = 1;
	}
	else if (strcmp(flagName, "IgnoreEntities") == 0)
	{
		ParserSetFlag(self->parser, IgnoreEntities, flag);
		self->parserFlags.IgnoreEntities = flag;
		result = 1;
	}
	else if (strcmp(flagName, "XMLLessThan") == 0)
	{
		ParserSetFlag(self->parser, XMLLessThan, flag);
		self->parserFlags.XMLLessThan = flag;
		result = 1;
	}
	else if (strcmp(flagName, "IgnorePlacementErrors") == 0)
	{
		ParserSetFlag(self->parser, IgnorePlacementErrors, flag);
		self->parserFlags.IgnorePlacementErrors = flag;
		result = 1;
	}
	else if (strcmp(flagName, "Validate") == 0)
	{
		ParserSetFlag(self->parser, Validate, flag);
		self->parserFlags.Validate = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ErrorOnValidityErrors") == 0)
	{
		ParserSetFlag(self->parser, ErrorOnValidityErrors, flag);
		self->parserFlags.ErrorOnValidityErrors = flag;
		result = 1;
	}
	else if (strcmp(flagName, "XMLSpace") == 0)
	{
		ParserSetFlag(self->parser, XMLSpace, flag);
		self->parserFlags.XMLSpace = flag;
		result = 1;
	}
	else if (strcmp(flagName, "XMLNamespaces") == 0)
	{
		ParserSetFlag(self->parser, XMLNamespaces, flag);
		self->parserFlags.XMLNamespaces = flag;
		result = 1;
	}
	else if (strcmp(flagName, "NoNoDTDWarning") == 0)
	{
		ParserSetFlag(self->parser, NoNoDTDWarning, flag);
		self->parserFlags.NoNoDTDWarning = flag;
		result = 1;
	}
	else if (strcmp(flagName, "SimpleErrorFormat") == 0)
	{
		ParserSetFlag(self->parser, SimpleErrorFormat, flag);
		self->parserFlags.SimpleErrorFormat = flag;
		result = 1;
	}
	else if (strcmp(flagName, "AllowUndeclaredNSAttributes") == 0)
	{
		ParserSetFlag(self->parser, AllowUndeclaredNSAttributes, flag);
		self->parserFlags.AllowUndeclaredNSAttributes = flag;
		result = 1;
	}
	else if (strcmp(flagName, "RelaxedAny") == 0)
	{
		ParserSetFlag(self->parser, RelaxedAny, flag);
		self->parserFlags.RelaxedAny = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ReturnNamespaceAttributes") == 0)
	{
		ParserSetFlag(self->parser, ReturnNamespaceAttributes, flag);
		self->parserFlags.ReturnNamespaceAttributes = flag;
		result = 1;
	}
	else if (strcmp(flagName, "ProcessDTD") == 0)
	{
		ParserSetFlag(self->parser, ProcessDTD, flag);
		self->parserFlags.ProcessDTD = flag;
		result = 1;
	} /* if */

	return result;
} /* set_flag */

/* ------------------------------------------- */

static char Rxpparse_setParserFlag__doc__[] = "\
Activate (de-activate) an event method.\n\
Usage:\n\
    parser.setParserFlag(flagName, flag)\n\
Where:\n\
    flagName is one of the following:\n\
        ExpandCharacterEntities\n\
        ExpandGeneralEntities\n\
        XMLSyntax\n\
        XMLPredefinedEntities\n\
        ErrorOnUnquotedAttributeValues\n\
        NormaliseAttributeValues\n\
        ErrorOnBadCharacterEntities\n\
        ErrorOnUndefinedEntities\n\
        ReturnComments\n\
        CaseInsensitive\n\
        ErrorOnUndefinedElements\n\
        ErrorOnUndefinedAttributes\n\
        WarnOnRedefinitions\n\
        TrustSDD\n\
        XMLExternalIDs\n\
        ReturnDefaultedAttributes\n\
        MergePCData\n\
        XMLMiscWFErrors\n\
        XMLStrictWFErrors\n\
        AllowMultipleElements\n\
        MaintainElementStack\n\
        IgnoreEntities\n\
        XMLLessThan\n\
        IgnorePlacementErrors\n\
        Validate\n\
        ErrorOnValidityErrors\n\
        XMLSpace\n\
        XMLNamespaces\n\
        NoNoDTDWarning\n\
        SimpleErrorFormat\n\
        AllowUndeclaredNSAttributes\n\
        RelaxedAny\n\
        ReturnNamespaceAttributes\n\
        ProcessDTD\n\
    flag is 0 (clear) or 1 (set)\n\
Returns: 1 if successful.  0 if unsuccessful (e.g. no such flag).\n\
";
static PyObject *
Rxpparse_setParserFlag(RxpparseObject * self, PyObject *args)
{
	char * flagName;
	int flag;
	int result;
	PyObject * pyResult;

	if (!PyArg_ParseTuple(args, "si:setParserFlag", &flagName, &flag))
	{
		return NULL;
	} /* if */

	result = set_flag(self, flagName, flag);
	pyResult = PyInt_FromLong((long)result);
	return pyResult;
} /* Rxpparse_setParserFlag */

/* ------------------------------------------- */

static char Rxpparse_activateByNameAndMethod__doc__[] = "\
Activate (de-activateByNameAndMethod) an event method.\n\
Usage:\n\
    parser.activateByNameAndMethod(handlerName, flag)\n\
Where:\n\
    handlerName is 'startDocument', 'endDocument', 'characters', etc.\n\
    flag is 1 (activateByNameAndMethod) or 2 (de-activate)\n\
";
static PyObject *
Rxpparse_activateByNameAndMethod(RxpparseObject *self, PyObject *args)
{
	char * methodName;
	PyObject * method;
	int flag;
	int result;
	PyObject * pyResult;

	if (!PyArg_ParseTuple(args, "sOi:activateByNameAndMethod",
		&methodName, &method, &flag))
	{
		return NULL;
	} /* if */

	if (! PyMethod_Check(method))
	{
		PyErr_SetString(PyExc_TypeError, "handler  required");
		return NULL;
	} /* if */

	result = activate_methods(self, methodName, flag);
	pyResult = PyInt_FromLong((long)0);
	return pyResult;
} /* Rxpparse_activateByNameAndMethod */

/* ------------------------------------------- */

static PyMethodDef Rxpparse_methods[] = {
	{"parseFile",	(PyCFunction)Rxpparse_parseFile,
		METH_VARARGS, Rxpparse_parseFile__doc__},
	{"parse",	(PyCFunction)Rxpparse_parse,
		METH_VARARGS, Rxpparse_parse__doc__},
	{"setContentHandler",		(PyCFunction)Rxpparse_setContentHandler,
		METH_VARARGS, Rxpparse_setContentHandler__doc__},
	{"setDTDHandler",		(PyCFunction)Rxpparse_setDTDHandler,
		METH_VARARGS, Rxpparse_setDTDHandler__doc__},
	{"setErrorHandler",		(PyCFunction)Rxpparse_setErrorHandler,
		METH_VARARGS, Rxpparse_setErrorHandler__doc__},
	{"getLocation",		(PyCFunction)Rxpparse_getLocation,
		METH_VARARGS, Rxpparse_getLocation__doc__},
	{"activate",	(PyCFunction)Rxpparse_activate,
		METH_VARARGS, Rxpparse_activate__doc__},
	{"getParserFlag",	(PyCFunction)Rxpparse_getParserFlag,
		METH_VARARGS, Rxpparse_getParserFlag__doc__},
	{"setParserFlag",	(PyCFunction)Rxpparse_setParserFlag,
		METH_VARARGS, Rxpparse_setParserFlag__doc__},
	{"activateByNameAndMethod",	(PyCFunction)Rxpparse_activateByNameAndMethod,
		METH_VARARGS, Rxpparse_activateByNameAndMethod__doc__},
	{NULL,		NULL}		/* sentinel */
};

/* ------------------------------------------- */

static PyObject *
Rxpparse_getattr(RxpparseObject *self, char *name)
{
	if (self->x_attr != NULL)
	{
		PyObject *v = PyDict_GetItemString(self->x_attr, name);
		if (v != NULL)
		{
			Py_INCREF(v);
			return v;
		}
	}
	return Py_FindMethod(Rxpparse_methods, (PyObject *)self, name);
}

/* ------------------------------------------- */

static int
Rxpparse_setattr(RxpparseObject *self, char *name, PyObject *v)
{
	if (self->x_attr == NULL) {
		self->x_attr = PyDict_New();
		if (self->x_attr == NULL)
			return -1;
	}
	if (v == NULL)
	{
		int rv = PyDict_DelItemString(self->x_attr, name);
		if (rv < 0)
		{
			PyErr_SetString(PyExc_AttributeError,
			        "delete non-existing Rxpparse attribute");
		} /* if */
		return rv;
	}
	else
	{
		return PyDict_SetItemString(self->x_attr, name, v);
	} /* if */
}

/* ------------------------------------------- */

static char Rxpparse_Type_documentation[] = "\
A SAX parser object implemented on top of rxp.\n\
";

statichere PyTypeObject Rxpparse_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"rxpopmodule.Rxpparse",		/*tp_name*/
	sizeof(RxpparseObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)Rxpparse_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)Rxpparse_getattr, /*tp_getattr*/
	(setattrfunc)Rxpparse_setattr, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
        0,                      /*tp_call*/
        0,                      /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        Rxpparse_Type_documentation,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        0,                      /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        0,                      /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        0,                      /*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/
};

/* ------------------------------------------- */

static PyObject *
rxpop_new(PyObject *self, PyObject *args)
{
	RxpparseObject *rv;
	
	if (!PyArg_ParseTuple(args, ":new"))
		return NULL;
	rv = newRxpparseObject(args);
	if ( rv == NULL )
	    return NULL;
	return (PyObject *)rv;
}

/* ------------------------------------------- */

static char rxpop_ParserCreate__doc__[] = "\
Create a new Parser object.\n\
";
static PyObject *
rxpop_ParserCreate(PyObject *self, PyObject *args)
{
	RxpparseObject * obj;

	if (!PyArg_ParseTuple(args, ":ParserCreate"))
	{
		return NULL;
	} /* if */
	obj = newRxpparseObject(NULL);
	if (obj != NULL)
	{
		return (PyObject *)obj;
	} /* if */
	Py_INCREF(Py_None);
	return Py_None;
} /* rxpop_ParserCreate */

/* ------------------------------------------- */

static char rxpop_GetDriverVersion__doc__[] = "\
Return the driver version (a string).\n\
";
static PyObject *
rxpop_GetDriverVersion(PyObject *self, PyObject *args)
{
	PyObject * version;

	if (!PyArg_ParseTuple(args, ":GetDriverVersion"))
	{
		return NULL;
	} /* if */

	version = PyString_FromString("1.0a");
	return version;

} /* rxpop_GetDriverVersion */

/* ------------------------------------------- */

static char rxpop_ErrorString__doc__[] = "\
ErrorString(errno) -> string\n\
Return string error for given number.\n\
Usage: ErrorString(errno)\n\
";
static PyObject *
rxpop_ErrorString(PyObject *self, PyObject *args)
{
	PyObject *a;
	long b;
	if (!PyArg_ParseTuple(args, "O#:ErrorString", &a, &b))
		return NULL;
	Py_INCREF(Py_None);
	return Py_None;
}

/* ------------------------------------------- */

/* List of functions defined in the module */

static PyMethodDef rxpop_methods[] = {
	{"ParserCreate",		rxpop_ParserCreate,
		METH_VARARGS, rxpop_ParserCreate__doc__},
	{"GetDriverVersion",		rxpop_GetDriverVersion,
		METH_VARARGS, rxpop_GetDriverVersion__doc__},
	{"ErrorString",			rxpop_ErrorString,
		METH_VARARGS, rxpop_ErrorString__doc__},
	{"new",		rxpop_new,		METH_VARARGS},
	{NULL,		NULL}		/* sentinel */
};

/* ------------------------------------------- */

/* Module docstring
*/

static char rxp_module_documentation[] = 
"Python wrapper for rxp SAX parser.";

/* ------------------------------------------- */

/* Initialization function for the module (*must* be called initrxpop)
*/

DL_EXPORT(void)
initrxpop(void)
{
	PyObject *m, *d;

	/* Initialize the type of the new type object here; doing it here
	 * is required for portability to Windows without requiring C++. */
	Rxpparse_Type.ob_type = &PyType_Type;

	/* Create the module and add the functions */
	m = Py_InitModule3("rxpop", rxpop_methods, rxp_module_documentation);

	/* Add some symbolic constants to the module */
	d = PyModule_GetDict(m);
	ErrorObject = PyErr_NewException("rxpop.error", NULL, NULL);
	PyDict_SetItemString(d, "error", ErrorObject);
}

/* ------------------------------------------- */


